//package urdar;
import java.util.*;
import java.io.*;

/** The <code>World</code> class represents the world in which the simulation takes place.
 * The main components of the <code>World</code> object are the population of organisms and the resource pool. 
 * Both these are defined using the <code>Vector</code> class.
 * @author Philip Gerlee
 * @author Torbjorn Lundh
 * @version 0.1 1st Jan 2011
 */
public class World {
    /** 
     * A Vector containing all the organisms in the World
     */
    public Vector<Organism> organisms;
    /** 
     * A Vector containing all the bit strings in the World
     */
    public Vector<BString> resources;
    /** 
     * The length of the bit strings
     */
    public int stringLength;
    /** 
     * Counter for the number of organism divisions
     */
    public int noDivisions;
    /** 
     * The level of entropy calculation
     */
    public int entropyLevel;
    /** 
     * Counter for the number of bit strings flowing into the system
     */
    public double noStringsFlowing; 
    /** 
     * The flow rate of bit strings
     */
    public double flow;
    /** 
     * The bit flip probability of inflowing strings
     */
    public double bitflipProb; 
    /** 
     * In the case of fixed strings this is the inflowing string
     */
    public boolean[] fixedString;
    /** 
     * True if random bit strings are used
     */
    public boolean randomStrings;
    /** 
     * When strings are random this declares the type of strings being used
     */
    public int randomMode; 
    /** 
     * The matrix holding interactions between the species.  I(i,j) = the number of bit strings that have been metabolised by species i and then by species j
     */
    public int[][] interactions;
    /** 
     * An array that holds of the productivity for the last <code>avg</code> time steps
     */
    public double[] productivity;
    /** 
     * The number of updates the efficiency is averaged over
     */
    public static int avg = 20;
    /** 
     * The entropy of the inflowing strings
     */
    public double inEntropy; 

    /** Creates a World with a flow of random strings
     * @param popSize the number of agents in the population
     * @param foodSize the number of bit strings in the resource pool
     * @param stringLength the length of the bit strings
     * @param flow the flow rate of new strings into the system
     * @param bitflipProb the bitflip probability of the inflowing strings
     * @param randomMode the type of random strings flowing into the system
     * @param entropyLevel the block length at which the entropy is calculated 
     * @see BString#BString(int, double, int)
     */
    public World(int popSize, int foodSize, int stringLength, double flow, double bitflipProb, int randomMode,int entropyLevel) {
	organisms = new Vector<Organism>(popSize);
	resources = new Vector<BString>(foodSize);
	this.flow = flow;
	this.stringLength = stringLength;
	this.bitflipProb = bitflipProb;
	this.randomMode = randomMode;
	this.entropyLevel = entropyLevel;
	randomStrings = true;
	noDivisions = 0;
	noStringsFlowing = 0.0;
	interactions = new int[256][256];
	double meanEntropy = 0.0;
	for (int i = 0; i < 256; i++) {
	    for (int j = 0; j < 256; j++) {
		interactions[i][j]=0;
	    }
	}
	productivity = new double[avg];
	for (int i = 0; i < avg; i++) {
	    productivity[i] = 0.0;
	}

	//Initiating the organism population as a uniform distribution across all 256 species
	for (int i = 0; i < popSize; i++) {
	    organisms.add(i,new Organism(i%256,i,Organism.getNextGID()));
	    organisms.get(i).setString(new BString(stringLength,bitflipProb,randomMode));
	}
	//Initate food pool and calculate normalisation constant inEntropy for the Efficiency
	for (int i = 0; i < foodSize; i++) {
	    resources.add(i,new BString(stringLength,bitflipProb,randomMode));
	    meanEntropy = meanEntropy + resources.get(i).Entropy(entropyLevel);
	}
	inEntropy = meanEntropy/foodSize;
    }

    /** Creates a World with a flow of fixed strings
     * @param popSize the number of agents in the population
     * @param foodSize the number of bit strings in the resource pool
     * @param flow the flow rate of new strings into the system
     * @param fixedString the bit string that flows into the system
     * @param entropyLevel the block length at which the entropy is calculated 
     */
    public World(int popSize, int foodSize, double flow, boolean[] fixedString, int entropyLevel) {
	organisms = new Vector<Organism>(popSize);
	resources = new Vector<BString>(foodSize);
	this.flow = flow;
	this.fixedString = fixedString;
	stringLength = fixedString.length;
	bitflipProb = 0;
	noDivisions = 0;
	this.entropyLevel = entropyLevel;
	randomStrings = false;
	interactions = new int[256][256];
	double meanEntropy = 0.0;
	for (int i = 0; i < 256; i++) {
	    for (int j = 0; j < 256; j++) {
		interactions[i][j]=0;
	    }
	}
	productivity = new double[avg];
	for (int i = 0; i < avg; i++) {
	    productivity[i] = 0.0;
	}
		
	//Initiating the organism population as a uniform distribution across all 256 species
	for (int i = 0; i < popSize; i++) {
	    organisms.add(i,new Organism(i%256,i,Organism.getNextGID()));
	    organisms.get(i).setString(new BString(fixedString));
	}
	//Initate food pool and calculate normalisation constant inEntropy for the Efficiency
	for (int i = 0; i < foodSize; i++) {
	    resources.add(i,new BString(fixedString));
	    meanEntropy = meanEntropy + resources.get(i).Entropy(entropyLevel);
	}
	inEntropy = meanEntropy/foodSize;	
    }

    /**Setting the rule frequency to defined values
     * @param rules the rules with non-zero abundance
     * @param fracs the corresponding frequencies. Should sum to one.
     */
    public void Inoculate(int[] rules, double[] fracs) {
	int[] nums = new int[rules.length];
	int k = 0;
	//Converting fractions to numbers 
	for (int i = 0; i < nums.length; i++) {
	    nums[i] = (int) Math.floor(fracs[i]*organisms.size());
	}
	//Initiating the population
	for (int i = 0; i < rules.length; i++) {
	    for (int j = 0; j < nums[i]; j++) {
		organisms.set(k,new Organism(rules[i],k,Organism.getNextGID()));
		organisms.get(k).setString(new BString(stringLength,bitflipProb,randomMode));
		k++;
	    }
	}
    }

    /**The average entropy of bit strings
     * @return the average entropy of the strings in the resource pool
     */
    public double getAvgEntropy() {
	double S = 0.0;
	for (int i = 0; i < resources.size(); i++) {
	    S = S + resources.get(i).Entropy(entropyLevel);
	}
	return S/resources.size();
    }

    /**The average IO (entropy difference) of the organisms
     * @return the average IO
     */ 
    public double getAvgIO() {
	double S = 0.0;
	for (int i = 0; i < organisms.size(); i++) {
	    S = S + organisms.get(i).getLatestIO();
	}
	return S/organisms.size();
    }

    /**The average IO (entropy difference) of organisms of a certain genotype 
     * @param rule the genotype under consideration
     * @return the average IO
     */
    public double getIO(int rule) {
	double S = 0.0;
	int j = 0;
	for (int i = 0; i < organisms.size(); i++) {
	    //Picking out organisms of type $rule
	    if (organisms.get(i).getGenome()==rule) {
		S = S + organisms.get(i).getLatestIO();
		j++;
	    }
	}
	if (j > 0) {
	    return S/j;
	}
	else {
	    return 0;
	}
    }

    /**The average age of the organisms
     * @return the average age (in updates)
     */
    public double getAvgAge() {
	double S = 0.0;
	for (int i = 0; i < organisms.size(); i++) {
	    S = S + organisms.get(i).getAge();
	}
	return (double) S/organisms.size();
    }


    /**The average age of the bit strings. Here the age is defined as the number manipulations the bit string has been through
     * @return the average age
     */
    public double getAvgStringAge() {
	double S = 0.0;
	for (int i = 0; i < resources.size(); i++) {
	    //Lenght of the metabolic history corresponds to age
	    S = S + resources.get(i).getHistory().size();
	}
	return (double) S/resources.size();
    }

    /**The average amount of energy that has been extracted from the bit strings. This quantity is averaged over the last $avg updates
     * @return the average productivity
     */
    public double getAvgProd() {
	double prod = 0.0;
	for (int i = 1; i < avg; i++) {
	    prod = prod + productivity[i];
	}
	return prod/avg;
    }

    /**The average efficiency of the organisms, defined as the average IO divided by the maximum amount of energy that can be extracted from the strings
     * @return the average efficiency
     */
    public double getEff() {
	//Normalising the measured productvity with respect to maximum possible productivity
	return getAvgProd()/(flow*resources.size()*(1.0-inEntropy)); //Change
    }

    /**The total number of organism divisions that have occured in the system
     * @return the number of divisions
     */
    public int getDivides() {
	return noDivisions;
    }

    /**The number of genotypes/rules that exist above a certain threshold
     * @param frac the threshold
     * @return the number of genotypes
     */
    public int getNoPhenotypes(double frac) {
	int N = 256;
	double[] f = new double[256];
	for (int i = 0; i < f.length; i++) {
	    f[i] = 0;
	}
	for (int i = 0; i < organisms.size(); i++) {
	    f[organisms.get(i).getGenome()]++;
	}
	for (int i = 0; i < f.length; i++) {
	    if (f[i] < (double) frac*organisms.size()) {
		N--;
	    }
	}
	return N;
    }

    /**The fraction of the population that corresponds to a given rule
     * @param rule the rule
     * @return the fraction
     */
    public double getFraction(int rule) {
	int sum = 0;
	for (int i = 0; i < organisms.size(); i++) {
	    if (organisms.get(i).getGenome() == rule) {
		sum++;
	    }
	}
	return (double) sum/organisms.size();
    }

    /**Sets the level at which the entropy is calculated
     * @param x the level
     */
    public void setEntLevel(int x) {
	entropyLevel = x;
    }

    /**Sets the flow rate
     * @param fl the rate
     */
    public void setFlow(double fl) {
	flow = fl;
    }

    /**Sets the mutation rate
     * @param pc the mutation rate
     */
    public void setMutRate(double pc) {
	Organism.setMutRate(pc);
    }

    /**Adds an interaction to the matrix I
     * @param H the metabolic history of the bit string manifesting the interaction 
     */
    public void addInt(Vector<Integer> H) {
	int n = H.size();
	int x = H.get(n-1).intValue();
	int y = H.get(n-2).intValue();
	if (y > 0) {
	    interactions[x][y]=interactions[x][y]+1;  
	}
    }

    /**Updates all organisms in the population, takes care of the swapping of strings, and calls the divide method of Organism
     * @see Organism#divide()
     * @see World#swapString(BString)
     */
    public void update() { //This method iterates the dynamics of the simulation
	Organism child = null;
	int div = 0; //indicator for division
	int n = 0; //counter for the flow
	double Pr = 0.0; //counter productivity
	double E0 = 0.0; //keeps track of entropy of old strings
	double E1 = 0.0; //keeps track of entropy of new strings
	//Clearing newBorn-tags from all organisms
	for (int i = 0; i < organisms.size(); i++) {
	    organisms.get(i).clearTag();
	}
	//Updating all organisms
	for (int i = 0; i < organisms.size(); i++) {
	    div = organisms.get(i).update(entropyLevel);
	    if (div == 1) { //dE > 0 and division occurs
		addInt(organisms.get(i).getString().getHistory()); //Adding interaction
		child = organisms.get(i).divide(); //Perform the division
		organisms.get(i).setString(this.swapString(organisms.get(i).getString())); //Swapping string
		n = (int) Math.floor(organisms.size()*Math.random()); //Picking random org. to be replaced
		child.setString(organisms.get(n).getOrigString()); //Setting child string to string of replaced organism
		child.setString(this.swapString(child.getString())); //Getting a new string for the child
		organisms.set(n,child); //Placing child in population
		noDivisions++; //Increment divisions
	    }
	    else if (div == 2) { //dE > 0 but no division occurs
		organisms.get(i).setString(this.swapString(organisms.get(i).getString())); //Swapping string
		    
	    } else { //dE < 0 and consequently no division occurs
		//If no division occurs put the original bit string back into the resource pool
		organisms.get(i).setString(this.swapString(organisms.get(i).getOrigString()));
	    }
	}
	    
	//Performing the flow of new bit strings into the resource pool
	double Fsize = (double) resources.size();
	noStringsFlowing = noStringsFlowing + Fsize*flow;
	while (noStringsFlowing >= 1) {
	    noStringsFlowing--;
	    n = (int) Math.floor(resources.size()*Math.random()); //Picking random string to be replaced
	    E0 = resources.get(n).Entropy(entropyLevel);
	    if (randomStrings) {
		resources.set(n,new BString(stringLength,bitflipProb,randomMode));
		E1 = resources.get(n).Entropy(entropyLevel);
	    }
	    else {
		resources.set(n,new BString(fixedString));
		E1 = resources.get(n).Entropy(entropyLevel);
	    }
	    //Calculating difference between outgoing and incoming strings, i.e. the productivity of the organisms
	    Pr = Pr+(E0-E1);
	}
	for (int i = 1; i < avg; i++) {
	    productivity[i-1] = productivity[i];
	}
	productivity[avg-1] = Pr;
    }

    /**Updates all organisms in the population without carrying out the actual divisions
     * @see World#swapString(BString)
     */
    public void fakeUpdate() { //This method iterates the dynamics of the simulation
	BString S;
	int div = 0;
	int n = 0; //counter for the flow
	double Pr = 0.0; //counter productivity
	double E0 = 0.0; //keeps track of entropy of old strings
	double E1 = 0.0; //keeps track of entropy of new strings
	for (int i = 0; i < organisms.size(); i++) {
	    organisms.get(i).clearTag();
	}
	for (int i = 0; i < organisms.size(); i++) {
	    div = organisms.get(i).update(entropyLevel);
	    if (div > 0) {
		organisms.get(i).setString(this.swapString(organisms.get(i).getString())); //Swapping string
		noDivisions++;
	    }
	    else {
		organisms.get(i).setString(this.swapString(organisms.get(i).getOrigString())); //IO
	    }
	}

	double Fsize = (double) resources.size();
	noStringsFlowing = noStringsFlowing + Fsize*flow;
	while (noStringsFlowing >= 1) {
	    noStringsFlowing--;
	    n = (int) Math.floor(resources.size()*Math.random()); //Picking random string to be replaced
	    E0 = resources.get(n).Entropy(entropyLevel);
	    if (randomStrings) {
		resources.set(n,new BString(stringLength,bitflipProb,randomMode));
		E1 = resources.get(n).Entropy(entropyLevel);
	    }
	    else {
		resources.set(n,new BString(fixedString));
		E1 = resources.get(n).Entropy(entropyLevel);
	    }
	    //Calculating difference between outgoing and incoming strings, i.e. the productivity of the organisms
	    Pr = Pr+(E0-E1);
	}
	for (int i = 1; i < avg; i++) {
	    productivity[i-1] = productivity[i];
	}
	productivity[avg-1] = Pr;
    }
    
    /**Swaps the specified bit string with a random string from the resource pool
     * @param S the bit string to be swapped
     * @return the new bit string
     */
    public BString swapString(BString S) {
	BString St;
	int i = (int) Math.floor(resources.size()*Math.random());
	St = resources.get(i);
	resources.set(i,S);
	return St;
    }
    
    /**Saves all genotypes above a certain fraction to a data file parametrised by time
     * @param t the time, used for naming the data file
     * @param frac the threshold value
     */
    public void saveGenotypes(int t,double frac) throws IOException, FileNotFoundException { 
	FileOutputStream fos = new FileOutputStream("gtypes" + Integer.toString(t) + ".txt",false);
	PrintWriter w = new PrintWriter(fos);
	organisms.get(0);
	int max = Organism.getCurrentGID();
	int[] genotypes = new int[max+1];
	for (int i = 0; i < genotypes.length-1; i++) {
	    genotypes[i] = 0;
	}
	for (int i = 0; i < organisms.size(); i++) {
	    genotypes[organisms.get(i).getGID()] = genotypes[organisms.get(i).getGID()] + 1;
	}
	for (int j = 0; j < organisms.size(); j++) {
	    if ((double) genotypes[organisms.get(j).getGID()]/organisms.size() > frac) {
		w.println(organisms.get(j).getGenome());
		genotypes[organisms.get(j).getGID()] = 0;
	    }
	}
	w.close();
    }
    
    /**Saves the metablic history of all strings to a data file parametrised by time
     * @param t the time, used for naming the data file
     */
    public void saveFoodHistory(int t)  throws IOException, FileNotFoundException { 
	Vector<Integer> H;
	FileOutputStream fos = new FileOutputStream("fhist" + Integer.toString(t) + ".txt",true);
	PrintWriter w = new PrintWriter(fos);
	//Looping through all bit strings
	for (int i = 0; i < resources.size(); i++) {
	    H = resources.get(i).getHistory();
	    for (int j = 0; j < H.size(); j++) {
		w.print(H.get(j).intValue() + " ");
	    }
	    w.println();
	}
	w.close();
    }
    
    /**Saves the entropy of all strings to a data file parametrised by time
     * @param t the time, used for naming the data file
     */
    public void saveFoodEntropy(int t)  throws IOException, FileNotFoundException { 
	FileOutputStream fos = new FileOutputStream("fent" + Integer.toString(t) + ".txt",false);
	PrintWriter w = new PrintWriter(fos);
	for (int i = 0; i < resources.size(); i++) {
	    w.println(resources.get(i).Entropy(entropyLevel));
	}
	w.close();
    }

    /**Saves the no. of organisms in each species (rule) in a data file
     * @param file the name of the data file
     */
    public void saveState(String file) throws IOException, FileNotFoundException {
	FileOutputStream fos = new FileOutputStream(file,true);
	PrintWriter w = new PrintWriter(fos);
	int[] A = new int[256];
	for (int i = 0; i < A.length; i++) {
	    A[i] = 0;
	}
	for (int i = 0; i < organisms.size(); i++) {
	    A[organisms.get(i).getGenome()] = A[organisms.get(i).getGenome()] + 1;
	}
	for (int i = 1; i < A.length; i++) {
	    w.print(A[i] + " ");
	}
	w.println();
	w.close();
    }

    /**Returns the species distribution
     * @return an array containing the frequency of all 256 species
     */
    public double[] getState() {
	double[] A = new double[256];
	for (int i = 0; i < A.length; i++) {
	    A[i] = 0.0;
	}
	for (int i = 0; i < organisms.size(); i++) {
	    A[organisms.get(i).getGenome()] = A[organisms.get(i).getGenome()] + 1;
	}
	for (int i = 0; i < A.length; i++) {
	    A[i] = (double) A[i]/organisms.size();
	}
	return A;
    }

    /**Saves any information in a data file
     * @param file the name of the data file
     * @param stats the data that is being written
     */
    public void saveStats(String file, String stats) throws IOException, FileNotFoundException {
	FileOutputStream fos = new FileOutputStream(file,true);
	PrintWriter w = new PrintWriter(fos);
	w.println(stats);
	w.close();
    }

    /**Saves the interaction matrix <code>interactions</code> in a data file
     * @param file the name of the data file
     * @see World#addInt(Vector)
     */
    public void saveInts(String file) throws IOException, FileNotFoundException {
	FileOutputStream fos = new FileOutputStream(file,true);
	PrintWriter w = new PrintWriter(fos);
	for (int i = 0; i < 256; i++) {
	    for (int j = 0; j < 256; j++) {
		w.print(interactions[i][j] + " ");
	    }
	    w.println();
	}
	w.close();
    }

    /**Saves the average efficiency in a data file
     * @param file the name of the data file
     */
    public void saveEff(String file) throws IOException, FileNotFoundException {
	FileOutputStream fos = new FileOutputStream(file,true);
	PrintWriter w = new PrintWriter(fos);
	w.println(getEff());
	w.close();
		
    }
}

